# Copyright 2023 Vincent Jacques

import unittest
import os

from . import *


forbid_gpu = os.environ.get("LINCS_DEV_FORBID_GPU", "false") == "true"

# @todo Test using named parameters when calling the API (e.g. Criterion(name="Criterion name", ...), Problem(criteria=[...], ...))

class ProblemTestCase(unittest.TestCase):
    def test_init_empty(self):
        problem = Problem([], [])
        self.assertEqual(len(problem.criteria), 0)
        self.assertEqual(len(problem.ordered_categories), 0)

    def test_init_wrong_types(self):
        with self.assertRaises(TypeError):
            Problem()
        with self.assertRaises(TypeError):
            Problem([])
        with self.assertRaises(TypeError):
            Problem([0], [])
        with self.assertRaises(TypeError):
            Problem([], [0])
        with self.assertRaises(TypeError):
            Problem([], [], 0)

    def test_init_one_criterion(self):
        problem = Problem([Criterion("Criterion name", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1)], [])
        self.assertEqual(len(problem.criteria), 1)
        self.assertEqual(problem.criteria[0].name, "Criterion name")
        self.assertEqual(problem.criteria[0].value_type, Criterion.ValueType.real)
        self.assertEqual(problem.criteria[0].preference_direction, Criterion.PreferenceDirection.increasing)

    def test_assign_criterion_attributes(self):
        # @todo(Feature, when there are more values in ValueType, later) Use other value in constructor
        problem = Problem([Criterion("Wrong criterion", Criterion.ValueType.real, Criterion.PreferenceDirection.decreasing, 0, 1)], [])
        self.assertEqual(problem.criteria[0].value_type, Criterion.ValueType.real)
        self.assertEqual(problem.criteria[0].preference_direction, Criterion.PreferenceDirection.decreasing)
        problem.criteria[0].name = "Criterion name"
        problem.criteria[0].value_type = Criterion.ValueType.real
        problem.criteria[0].preference_direction = Criterion.PreferenceDirection.increasing
        self.assertEqual(problem.criteria[0].name, "Criterion name")
        self.assertEqual(problem.criteria[0].value_type, Criterion.ValueType.real)
        self.assertEqual(problem.criteria[0].preference_direction, Criterion.PreferenceDirection.increasing)

    def test_assign_criterion(self):
        problem = Problem([Criterion("Wrong criterion", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1)], [])
        problem.criteria[0] = Criterion("Criterion name", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1)
        self.assertEqual(problem.criteria[0].name, "Criterion name")

    def test_append_criterion(self):
        problem = Problem([], [])
        problem.criteria.append(Criterion("Criterion name", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1))
        self.assertEqual(len(problem.criteria), 1)

    def test_assign_criteria_slice(self):
        problem = Problem([], [])
        problem.criteria[:] = [Criterion("Criterion name", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1)]
        self.assertEqual(len(problem.criteria), 1)

    def test_init_one_category(self):
        problem = Problem([], [Category("Category name")])
        self.assertEqual(len(problem.ordered_categories), 1)
        self.assertEqual(problem.ordered_categories[0].name, "Category name")

    def test_assign_category_attributes(self):
        problem = Problem([], [Category("Wrong category")])
        problem.ordered_categories[0].name = "Category name"
        self.assertEqual(problem.ordered_categories[0].name, "Category name")

    def test_assign_category(self):
        problem = Problem([], [Category("Wrong category")])
        problem.ordered_categories[0] = Category("Category name")
        self.assertEqual(problem.ordered_categories[0].name, "Category name")

    def test_append_category(self):
        problem = Problem([], [])
        problem.ordered_categories.append(Category("Category name"))
        self.assertEqual(len(problem.ordered_categories), 1)

    def test_assign_categories_slice(self):
        problem = Problem([], [])
        problem.ordered_categories[:] = [Category("Category name")]
        self.assertEqual(len(problem.ordered_categories), 1)

    def test_iso_antitone(self):
        self.assertEqual(Criterion.PreferenceDirection.isotone, Criterion.PreferenceDirection.increasing)
        self.assertEqual(Criterion.PreferenceDirection.antitone, Criterion.PreferenceDirection.decreasing)


class ModelTestCase(unittest.TestCase):
    def test_init_empty(self):
        problem = Problem([], [])
        model = Model(problem, [])
        self.assertEqual(len(model.boundaries), 0)

    def test_init_wrong_types(self):
        problem = Problem([], [])
        with self.assertRaises(TypeError):
            Model()
        with self.assertRaises(TypeError):
            Model(problem)
        with self.assertRaises(TypeError):
            Model(problem, 0)
        with self.assertRaises(TypeError):
            Model(0, [])
        with self.assertRaises(TypeError):
            Model(problem, [0])

    def test_init_one_empty_boundary(self):
        problem = Problem([], [])
        model = Model(problem, [Model.Boundary([], SufficientCoalitions(SufficientCoalitions.weights, []))])
        self.assertEqual(len(model.boundaries), 1)
        self.assertEqual(len(model.boundaries[0].profile), 0)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.kind, SufficientCoalitions.Kind.weights)
        self.assertEqual(len(model.boundaries[0].sufficient_coalitions.criterion_weights), 0)

    def test_init_three_criteria_two_categories_weights_boundary(self):
        problem = Problem(
            [
                Criterion("Criterion 1", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                Criterion("Criterion 2", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                Criterion("Criterion 3", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
            ], [
                Category("Category 1"),
                Category("Category 2"),
            ],
        )
        model = Model(
            problem,
            [
                Model.Boundary(
                    [5., 5., 5],
                    SufficientCoalitions(SufficientCoalitions.weights, [0.5, 0.25, 1])
                ),
            ],
        )
        self.assertEqual(len(model.boundaries), 1)
        self.assertEqual(len(model.boundaries[0].profile), 3)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.kind, SufficientCoalitions.Kind.weights)
        self.assertEqual(len(model.boundaries[0].sufficient_coalitions.criterion_weights), 3)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.criterion_weights[0], 0.5)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.criterion_weights[1], 0.25)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.criterion_weights[2], 1)
        self.assertEqual(len(model.boundaries[0].sufficient_coalitions.upset_roots), 0)

    def test_init_three_criteria_two_categories_roots_boundary(self):
        problem = Problem(
            [
                Criterion("Criterion 1", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                Criterion("Criterion 2", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                Criterion("Criterion 3", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
            ], [
                Category("Category 1"),
                Category("Category 2"),
            ],
        )
        model = Model(
            problem,
            [
                Model.Boundary(
                    [5., 5., 5],
                    SufficientCoalitions(SufficientCoalitions.roots, 3, [[0, 1], [0, 2]])
                ),
            ],
        )
        self.assertEqual(len(model.boundaries), 1)
        self.assertEqual(len(model.boundaries[0].profile), 3)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.kind, SufficientCoalitions.Kind.roots)
        self.assertEqual(len(model.boundaries[0].sufficient_coalitions.criterion_weights), 0)
        self.assertEqual(len(model.boundaries[0].sufficient_coalitions.upset_roots), 2)
        self.assertEqual(len(model.boundaries[0].sufficient_coalitions.upset_roots[0]), 2)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.upset_roots[0][0], 0)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.upset_roots[0][1], 1)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.upset_roots[1][0], 0)
        self.assertEqual(model.boundaries[0].sufficient_coalitions.upset_roots[1][1], 2)

    def test_assign_model_attributes(self):
        problem = Problem([], [])
        model = Model(problem, [])
        model.boundaries = [Model.Boundary([], SufficientCoalitions(SufficientCoalitions.weights, []))]
        self.assertEqual(len(model.boundaries), 1)


class AlternativesTestCase(unittest.TestCase):
    def test_init_empty(self):
        problem = Problem([], [])
        alternatives = Alternatives(problem, [])
        self.assertEqual(len(alternatives.alternatives), 0)

    def test_init_wrong_types(self):
        problem = Problem([], [])
        with self.assertRaises(TypeError):
            Alternatives()
        with self.assertRaises(TypeError):
            Alternatives(problem)
        with self.assertRaises(TypeError):
            Alternatives(problem, 0)
        with self.assertRaises(TypeError):
            Alternatives(0, [])
        with self.assertRaises(TypeError):
            Alternatives(problem, [0])

    def test_init_three_criteria_two_categories(self):
        problem = Problem(
            [
                Criterion("Criterion 1", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                Criterion("Criterion 2", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                Criterion("Criterion 3", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
            ], [
                Category("Category 1"),
                Category("Category 2"),
            ],
        )
        alternatives = Alternatives(
            problem,
            [
                Alternative("First alternative", [5., 5., 5], 0),
                Alternative("Second alternative", [1., 2., 3.], None),
                Alternative("Third alternative", [2., 4., 6.]),
            ],
        )
        self.assertEqual(len(alternatives.alternatives), 3)

        self.assertEqual(alternatives.alternatives[0].name, "First alternative")
        self.assertEqual(list(alternatives.alternatives[0].profile), [5., 5., 5.])
        self.assertEqual(alternatives.alternatives[0].category_index, 0)

        self.assertEqual(alternatives.alternatives[1].name, "Second alternative")
        self.assertEqual(list(alternatives.alternatives[1].profile), [1., 2., 3.])
        self.assertIsNone(alternatives.alternatives[1].category_index)

        self.assertEqual(alternatives.alternatives[2].name, "Third alternative")
        self.assertEqual(list(alternatives.alternatives[2].profile), [2., 4., 6.])
        self.assertIsNone(alternatives.alternatives[2].category_index)


class LearningTestCase(unittest.TestCase):
    def test_basic_mrsort_learning(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
        weights_optimization_strategy = OptimizeWeightsUsingGlop(learning_data)
        profiles_improvement_strategy = ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
        breeding_strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, 4)
        termination_strategy = TerminateAtAccuracy(learning_data, len(learning_set.alternatives))
        learned_model = LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
        ).perform()

        self.assertEqual(learning_data.iteration_index, 8)

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 29)
        self.assertEqual(result.unchanged, 971)

    def test_iterations_restricted_mrsort_learning(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 1000, 43)

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
        weights_optimization_strategy = OptimizeWeightsUsingGlop(learning_data)
        profiles_improvement_strategy = ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
        breeding_strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, 4)
        termination_strategy = TerminateAfterIterations(learning_data, 1)
        learned_model = LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
        ).perform()

        self.assertGreater(classify_alternatives(problem, learned_model, learning_set).changed, 0)

    def test_terminate_when_any(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        class MyTerminationStrategy(LearnMrsortByWeightsProfilesBreed.TerminationStrategy):
            def __init__(self):
                super().__init__()
                self.called_count = 0

            def terminate(self):
                self.called_count += 1
                return self.called_count == 6

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
        weights_optimization_strategy = OptimizeWeightsUsingGlop(learning_data)
        profiles_improvement_strategy = ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
        breeding_strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, 4)
        my_termination_strategy = MyTerminationStrategy()
        termination_strategy = TerminateWhenAny([my_termination_strategy, TerminateAtAccuracy(learning_data, len(learning_set.alternatives))])
        learned_model = LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
        ).perform()

        self.assertEqual(my_termination_strategy.called_count, 6)
        self.assertEqual(classify_alternatives(problem, learned_model, learning_set).changed, 6)

    def test_python_strategies(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        class MyProfileInitializationStrategy(LearnMrsortByWeightsProfilesBreed.ProfilesInitializationStrategy):
            def __init__(self, learning_data):
                super().__init__()
                self.strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
                self.called_count = 0

            def initialize_profiles(self, begin, end):
                self.called_count += 1
                return self.strategy.initialize_profiles(begin, end)

        class MyWeightsOptimizationStrategy(LearnMrsortByWeightsProfilesBreed.WeightsOptimizationStrategy):
            def __init__(self, learning_data):
                super().__init__()
                self.strategy = OptimizeWeightsUsingGlop(learning_data)
                self.called_count = 0

            def optimize_weights(self):
                self.called_count += 1
                return self.strategy.optimize_weights()

        class MyProfilesImprovementStrategy(LearnMrsortByWeightsProfilesBreed.ProfilesImprovementStrategy):
            def __init__(self, learning_data):
                super().__init__()
                self.strategy = ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
                self.called_count = 0

            def improve_profiles(self):
                self.called_count += 1
                return self.strategy.improve_profiles()

        class MyBreedingStrategy(LearnMrsortByWeightsProfilesBreed.BreedingStrategy):
            def __init__(self, learning_data, profiles_initialization_strategy, count):
                super().__init__()
                self.strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, count)
                self.called_count = 0

            def breed(self):
                self.called_count += 1
                return self.strategy.breed()

        class MyTerminationStrategy(LearnMrsortByWeightsProfilesBreed.TerminationStrategy):
            def __init__(self, learning_data):
                super().__init__()
                self.learning_data = learning_data
                self.accuracies = []

            def terminate(self):
                self.accuracies.append(learning_data.get_best_accuracy())
                return len(self.accuracies) == 2

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = MyProfileInitializationStrategy(learning_data)
        weights_optimization_strategy = MyWeightsOptimizationStrategy(learning_data)
        profiles_improvement_strategy = MyProfilesImprovementStrategy(learning_data)
        breeding_strategy = MyBreedingStrategy(learning_data, profiles_initialization_strategy, 4)
        termination_strategy = MyTerminationStrategy(learning_data)
        learned_model = LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
        ).perform()

        self.assertEqual(profiles_initialization_strategy.called_count, 2)
        self.assertEqual(weights_optimization_strategy.called_count, 2)
        self.assertEqual(profiles_improvement_strategy.called_count, 2)
        self.assertEqual(breeding_strategy.called_count, 1)
        self.assertEqual(termination_strategy.accuracies, [182, 192])

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 8)
        self.assertEqual(result.unchanged, 192)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 68)
        self.assertEqual(result.unchanged, 932)

    def test_observers(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        class MyObserver(LearnMrsortByWeightsProfilesBreed.Observer):
            def __init__(self, learning_data):
                super().__init__()
                self.learning_data = learning_data
                self.best_accuracies = []
                self.final_accuracy = None

            def after_iteration(self):
                self.best_accuracies.append(self.learning_data.get_best_accuracy())

            def before_return(self):
                self.final_accuracy = self.learning_data.get_best_accuracy()

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
        weights_optimization_strategy = OptimizeWeightsUsingGlop(learning_data)
        profiles_improvement_strategy = ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
        breeding_strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, 4)
        termination_strategy = TerminateAtAccuracy(learning_data, len(learning_set.alternatives))
        observer = MyObserver(learning_data)
        LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
            [observer],
        ).perform()

        self.assertEqual(observer.best_accuracies, [182, 192, 192, 193, 193, 194, 193, 199])
        self.assertEqual(observer.final_accuracy, 200)

    def test_alglib_mrsort_learning(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
        weights_optimization_strategy = OptimizeWeightsUsingAlglib(learning_data)
        profiles_improvement_strategy = ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
        breeding_strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, 4)
        termination_strategy = TerminateAtAccuracy(learning_data, len(learning_set.alternatives))
        learned_model = LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
        ).perform()

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 21)
        self.assertEqual(result.unchanged, 979)

    @unittest.skipIf(forbid_gpu, "Can't use GPU")
    def test_gpu_mrsort_learning(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learning_data = LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, 9, 44)
        profiles_initialization_strategy = InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)
        weights_optimization_strategy = OptimizeWeightsUsingGlop(learning_data)
        profiles_improvement_strategy = ImproveProfilesWithAccuracyHeuristicOnGpu(learning_data)
        breeding_strategy = ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, 4)
        termination_strategy = TerminateAtAccuracy(learning_data, len(learning_set.alternatives))
        learned_model = LearnMrsortByWeightsProfilesBreed(
            learning_data,
            profiles_initialization_strategy,
            weights_optimization_strategy,
            profiles_improvement_strategy,
            breeding_strategy,
            termination_strategy,
        ).perform()

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 29)
        self.assertEqual(result.unchanged, 971)

    def test_sat_by_coalitions_using_minisat_learning(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learned_model = LearnUcncsBySatByCoalitionsUsingMinisat(problem, learning_set).perform()

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 21)
        self.assertEqual(result.unchanged, 979)

    def test_sat_by_separation_using_minisat_learning(self):
        problem = generate_classification_problem(5, 2, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learned_model = LearnUcncsBySatBySeparationUsingMinisat(problem, learning_set).perform()

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 24)
        self.assertEqual(result.unchanged, 976)

    def test_max_sat_by_coalitions_using_evalmaxsat_learning(self):
        problem = generate_classification_problem(5, 3, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learned_model = LearnUcncsByMaxSatByCoalitionsUsingEvalmaxsat(problem, learning_set).perform()

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 28)
        self.assertEqual(result.unchanged, 972)

    def test_max_sat_by_separation_using_evalmaxsat_learning(self):
        problem = generate_classification_problem(5, 2, 41)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 200, 43)

        learned_model = LearnUcncsByMaxSatBySeparationUsingEvalmaxsat(problem, learning_set).perform()

        result = classify_alternatives(problem, learned_model, learning_set)
        self.assertEqual(result.changed, 0)
        self.assertEqual(result.unchanged, 200)

        testing_set = generate_classified_alternatives(problem, model, 1000, 44)
        result = classify_alternatives(problem, learned_model, testing_set)
        self.assertEqual(result.changed, 24)
        self.assertEqual(result.unchanged, 976)

    def test_learning_failure_exception(self):
        problem = generate_classification_problem(2, 2, 42)
        model = generate_mrsort_classification_model(problem, 42)
        learning_set = generate_classified_alternatives(problem, model, 100, 42)
        misclassify_alternatives(problem, learning_set, 10, 42 + 27)

        learning = LearnUcncsBySatByCoalitionsUsingMinisat(problem, learning_set)

        with self.assertRaises(LearningFailureException):
            learned_model = learning.perform()



# Copyright 2023 Vincent Jacques

import matplotlib.pyplot as plt


def visualize_model(problem, model, alternatives, alternatives_count, out):
    fig, ax = plt.subplots(1, 1, figsize=(6, 4), layout="constrained")

    xs = [criterion.name for criterion in problem.criteria]
    ys = [
        normalize_profile(problem.criteria, boundary.profile)
        for boundary in model.boundaries
    ]
    ys.append([1] * len(xs))
    unstacked_ys = [ys[0]]
    for ys1, ys2 in zip(ys[1:], ys[:-1]):
        unstacked_ys.append([y1 - y2 for y1, y2 in zip(ys1, ys2)])
    collections = ax.stackplot(
        xs, unstacked_ys,
        labels=[category.name for category in problem.ordered_categories],
        alpha=0.4,
    )
    colors = [collection.get_facecolor() for collection in collections]

    if alternatives:
        for alternative in alternatives.alternatives[:alternatives_count]:
            if alternative.category_index is None:
                color = "black"
            else:
                color = colors[alternative.category_index]
            ax.plot(
                xs, normalize_profile(problem.criteria, alternative.profile),
                "o--",
                label=alternative.name,
                color=color,
                alpha=1,
            )

    ax.legend()
    ax.set_ylim(-0.05, 1.05)
    ax.set_yticks([0, 1])
    ax.set_yticklabels(["worst", "best"])

    for y in [0, 1]:
        ax.axhline(y=y, color="black", alpha=0.2)
    for x in xs:
        ax.axvline(x=x, color="black", alpha=0.2)

    fig.savefig(out, format="png", dpi=100)
    plt.close(fig)


def normalize_profile(criteria, ys):
    return [
        normalize_value(criterion, y)
        for (criterion, y) in zip(criteria, ys)
    ]

def normalize_value(criterion, y):
    y = (y - criterion.min_value) / (criterion.max_value - criterion.min_value)
    if criterion.preference_direction == criterion.PreferenceDirection.decreasing:
        return 1 - y
    else:
        return y



import unittest

from . import Problem, Criterion, Category, Model, SufficientCoalitions


def describe_problem(problem: Problem):
    categories_count = len(problem.ordered_categories)
    assert categories_count >= 2
    criteria_count = len(problem.criteria)
    assert criteria_count >= 1

    category_names = [f'"{category.name}"' for category in problem.ordered_categories]
    category_names_joined = ", ".join(category_names[:-1]) + " and " + category_names[-1]
    yield f"This a classification problem into {categories_count} ordered categories named {category_names_joined}."
    yield f"The best category is {category_names[-1]} and the worst category is {category_names[0]}."

    if criteria_count == 1:
        yield "There is 1 classification criterion."
    else:
        yield f"There are {criteria_count} classification criteria (in no particular order)."
    for criterion in problem.criteria:
        yield f'Criterion "{criterion.name}" takes {criterion.value_type.name} values between {criterion.min_value:.1f} and {criterion.max_value:.1f} included.'
        if criterion.preference_direction == criterion.PreferenceDirection.increasing:
            yield f'Higher values of "{criterion.name}" are known to be better.'
        else:
            yield f'Lower values of "{criterion.name}" are known to be better.'


class DescribeProblemTestCase(unittest.TestCase):
    def _test(self, problem, expected):
        self.assertEqual(list(describe_problem(problem)), expected)

    def test_simplest(self):
        self._test(
            Problem(
                [
                    Criterion("Criterion", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                ],
                [Category("Bad"), Category("Good")],
            ),
            [
                'This a classification problem into 2 ordered categories named "Bad" and "Good".',
                'The best category is "Good" and the worst category is "Bad".',
                'There is 1 classification criterion.',
                'Criterion "Criterion" takes real values between 0.0 and 1.0 included.',
                'Higher values of "Criterion" are known to be better.'
            ]
        )

    def test_many_categories(self):
        self._test(
            Problem(
                [
                    Criterion("Criterion", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
                ],
                [Category("Worsestest"), Category("Interm 1"), Category("Interm 2"), Category("Interm 3"), Category("Bestestest")],
            ),
            [
                'This a classification problem into 5 ordered categories named "Worsestest", "Interm 1", "Interm 2", "Interm 3" and "Bestestest".',
                'The best category is "Bestestest" and the worst category is "Worsestest".',
                'There is 1 classification criterion.',
                'Criterion "Criterion" takes real values between 0.0 and 1.0 included.',
                'Higher values of "Criterion" are known to be better.',
            ]
        )

    def test_criteria_variety(self):
        self._test(
            Problem(
                [
                    Criterion("Increasing criterion", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, -5.2, 10.3),
                    Criterion("Decreasing criterion", Criterion.ValueType.real, Criterion.PreferenceDirection.decreasing, 5, 15),
                ],
                [Category("Bad"), Category("Good")],
            ),
            [
                'This a classification problem into 2 ordered categories named "Bad" and "Good".',
                'The best category is "Good" and the worst category is "Bad".',
                'There are 2 classification criteria (in no particular order).',
                'Criterion "Increasing criterion" takes real values between -5.2 and 10.3 included.',
                'Higher values of "Increasing criterion" are known to be better.',
                'Criterion "Decreasing criterion" takes real values between 5.0 and 15.0 included.',
                'Lower values of "Decreasing criterion" are known to be better.',
            ]
        )


def describe_model(problem: Problem, model: Model):
    categories_count = len(problem.ordered_categories)
    assert categories_count >= 2
    assert len(model.boundaries) == categories_count - 1
    criteria_count = len(problem.criteria)
    assert criteria_count >= 1

    def comma_and(s):
        s = list(s)
        if len(s) == 1:
            return s[0]
        elif len(s) == 2:
            return " and ".join(s)
        else:
            return ", ".join(s[:-1]) + ", and " + s[-1]  # https://en.wikipedia.org/wiki/Serial_comma

    def make_coalitions(boundary):
        assert boundary.sufficient_coalitions.kind == SufficientCoalitions.Kind.roots
        for coalition in boundary.sufficient_coalitions.upset_roots:
            criterion_names = []
            for criterion_index in coalition:
                criterion = problem.criteria[criterion_index]
                criterion_names.append(f'"{criterion.name}"')
            yield f'  - {comma_and(criterion_names)}'

    def make_profile(boundary):
        assert len(boundary.profile) == criteria_count
        for criterion, limit in zip(problem.criteria, boundary.profile):
            constraint = "at least" if criterion.preference_direction == criterion.PreferenceDirection.increasing else "at most"
            yield f'{constraint} {limit:.2f} on criterion "{criterion.name}"'

    is_uc = all(
        # @todo Provide equality operator (on the C++ side?).
        boundary.sufficient_coalitions.kind == model.boundaries[0].sufficient_coalitions.kind
        and list(boundary.sufficient_coalitions.criterion_weights) == list(model.boundaries[0].sufficient_coalitions.criterion_weights)
        and list(boundary.sufficient_coalitions.upset_roots) == list(model.boundaries[0].sufficient_coalitions.upset_roots)
        for boundary in model.boundaries
    )
    if is_uc:
        first_boundary = model.boundaries[0]
        if first_boundary.sufficient_coalitions.kind == SufficientCoalitions.Kind.weights:
            yield "This is a MR-Sort (a.k.a. 1-Uc-NCS) model: an NCS model where the sufficient coalitions are specified using the same criterion weights for all boundaries."
            yield "The weights associated to each criterion are:"
            assert len(first_boundary.sufficient_coalitions.criterion_weights) == criteria_count
            for criterion, weight in zip(problem.criteria, first_boundary.sufficient_coalitions.criterion_weights):
                yield f'  - Criterion "{criterion.name}": {weight:.2f}'
            yield "To get into an upper category, an alternative must be better than the following profiles on a set of criteria whose weights add up to at least 1:"
        else:
            assert first_boundary.sufficient_coalitions.kind == SufficientCoalitions.Kind.roots
            yield "This is a Uc-NCS model: an NCS model with the same sufficient coalitions for all boundaries."
            yield "The sufficient coalitions of criteria are the following, as well as any of their unions:"
            yield from make_coalitions(first_boundary)
            yield "To get into an upper category, an alternative must be better than the following profiles on a sufficient coalition of criteria:"
        for category, boundary in zip(problem.ordered_categories[1:], model.boundaries):
            yield f'  - For category "{category.name}": {comma_and(make_profile(boundary))}'
    else:
        yield "This is a generic NCS model; sufficient coalitions are specified for each boundary."
        for category, boundary in zip(problem.ordered_categories[1:], model.boundaries):
            if boundary.sufficient_coalitions.kind == SufficientCoalitions.Kind.weights:
                yield f'To get into category "{category.name}", an alternative must be better than the following profile on a set of criteria whose weights add up to at least 1:'
                for profile, weight in zip(make_profile(boundary), boundary.sufficient_coalitions.criterion_weights):
                    yield f'  - {profile} (weight: {weight:.2f})'
            else:
                assert boundary.sufficient_coalitions.kind == SufficientCoalitions.Kind.roots
                yield f'The sufficient coalitions for category "{category.name}" are the following, as well as any of their unions:'
                yield from make_coalitions(boundary)
                yield f'To get into category "{category.name}", an alternative must be better than the following profile on a sufficient coalition of criteria: {comma_and(make_profile(boundary))}'


class DescribeModelTestCase(unittest.TestCase):
    maxDiff = None

    problem = Problem(
        [
            Criterion("Criterion 1", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
            Criterion("Criterion 2", Criterion.ValueType.real, Criterion.PreferenceDirection.decreasing, 0, 1),
            Criterion("Criterion 3", Criterion.ValueType.real, Criterion.PreferenceDirection.increasing, 0, 1),
            Criterion("Criterion 4", Criterion.ValueType.real, Criterion.PreferenceDirection.decreasing, 0, 1),
        ],
        [
            Category("Bad"),
            Category("Intermediate"),
            Category("Good"),
        ],
    )

    def _test(self, model, expected):
        self.assertEqual(list(describe_model(self.problem, model)), expected)

    def test_mrsort(self):
        self._test(
            Model(
                self.problem,
                [
                    Model.Boundary([0.2, 0.8, 0.4, 0.7], SufficientCoalitions(SufficientCoalitions.weights, [0.7, 0.5, 0.4, 0.2])),
                    Model.Boundary([0.7, 0.7, 0.5, 0.3], SufficientCoalitions(SufficientCoalitions.weights, [0.7, 0.5, 0.4, 0.2])),
                ],
            ),
            [
                'This is a MR-Sort (a.k.a. 1-Uc-NCS) model: an NCS model where the sufficient coalitions are specified using the same criterion weights for all boundaries.',
                'The weights associated to each criterion are:',
                '  - Criterion "Criterion 1": 0.70',
                '  - Criterion "Criterion 2": 0.50',
                '  - Criterion "Criterion 3": 0.40',
                '  - Criterion "Criterion 4": 0.20',
                'To get into an upper category, an alternative must be better than the following profiles on a set of criteria whose weights add up to at least 1:',
                '  - For category "Intermediate": at least 0.20 on criterion "Criterion 1", at most 0.80 on criterion "Criterion 2", at least 0.40 on criterion "Criterion 3", and at most 0.70 on criterion "Criterion 4"',
                '  - For category "Good": at least 0.70 on criterion "Criterion 1", at most 0.70 on criterion "Criterion 2", at least 0.50 on criterion "Criterion 3", and at most 0.30 on criterion "Criterion 4"'
            ],
        )

    def test_ucncs(self):
        self._test(
            Model(
                self.problem,
                [
                    Model.Boundary([0.2, 0.8, 0.4, 0.7], SufficientCoalitions(SufficientCoalitions.roots, 4, [[0, 1], [0, 2], [1, 2, 3]])),
                    Model.Boundary([0.7, 0.7, 0.5, 0.3], SufficientCoalitions(SufficientCoalitions.roots, 4, [[0, 1], [0, 2], [1, 2, 3]])),
                ],
            ),
            [
                'This is a Uc-NCS model: an NCS model with the same sufficient coalitions for all boundaries.',
                'The sufficient coalitions of criteria are the following, as well as any of their unions:',
                '  - "Criterion 1" and "Criterion 2"',
                '  - "Criterion 1" and "Criterion 3"',
                '  - "Criterion 2", "Criterion 3", and "Criterion 4"',
                'To get into an upper category, an alternative must be better than the following profiles on a sufficient coalition of criteria:',
                '  - For category "Intermediate": at least 0.20 on criterion "Criterion 1", at most 0.80 on criterion "Criterion 2", at least 0.40 on criterion "Criterion 3", and at most 0.70 on criterion "Criterion 4"',
                '  - For category "Good": at least 0.70 on criterion "Criterion 1", at most 0.70 on criterion "Criterion 2", at least 0.50 on criterion "Criterion 3", and at most 0.30 on criterion "Criterion 4"'
            ],
        )

    def test_mixed(self):
        self._test(
            Model(
                self.problem,
                [
                    Model.Boundary([0.2, 0.8, 0.4, 0.7], SufficientCoalitions(SufficientCoalitions.roots, 4, [[0, 1], [0, 2], [1, 2, 3]])),
                    Model.Boundary([0.7, 0.7, 0.5, 0.3], SufficientCoalitions(SufficientCoalitions.weights, [0.7, 0.5, 0.4, 0.2])),
                ],
            ),
            [
                'This is a generic NCS model; sufficient coalitions are specified for each boundary.',
                'The sufficient coalitions for category "Intermediate" are the following, as well as any of their unions:',
                '  - "Criterion 1" and "Criterion 2"',
                '  - "Criterion 1" and "Criterion 3"',
                '  - "Criterion 2", "Criterion 3", and "Criterion 4"',
                'To get into category "Intermediate", an alternative must be better than the following profile on a sufficient coalition of criteria: at least 0.20 on criterion "Criterion 1", at most 0.80 on criterion "Criterion 2", at least 0.40 on criterion "Criterion 3", and at most 0.70 on criterion "Criterion 4"',
                'To get into category "Good", an alternative must be better than the following profile on a set of criteria whose weights add up to at least 1:',
                '  - at least 0.70 on criterion "Criterion 1" (weight: 0.70)',
                '  - at most 0.70 on criterion "Criterion 2" (weight: 0.50)',
                '  - at least 0.50 on criterion "Criterion 3" (weight: 0.40)',
                '  - at most 0.30 on criterion "Criterion 4" (weight: 0.20)'
            ]
        )



# Copyright 2023 Vincent Jacques

import glob
import itertools
import os
import setuptools
import setuptools.command.build_ext
import shutil
import subprocess
import sys


with open("lincs/__init__.py") as f:
    for line in f.readlines():
        if line.startswith("__version__ = "):
            version = line[15:-2]
            break
    else:
        assert False, "Version not found in lincs/__init__.py"


with open("README.rst") as f:
    long_description = f.read()
for file in ["COPYING", "COPYING.LESSER"]:
    long_description = long_description.replace(f" <{file}>`_", f" <https://github.com/MICS-Lab/lincs/blob/v{version}/{file}>`_")
for lang in ["yaml", "shell", "text", "diff"]:
    long_description = long_description.replace(f".. highlight:: {lang}", "")


with open("requirements.txt") as f:
    install_requires = f.readlines()


windows_cuda_path = os.environ.get("CUDA_PATH", r"C:\Program Files\NVIDIA GPU Computing Toolkit\CUDA\v12.2")


# Method for building an extension with CUDA code extracted from https://stackoverflow.com/a/13300714/905845
# @todo(Project management, later) Consider using scikit-build:
# it should make it easier to compile CUDA code using nvcc and to run C++ unit tests during build.
# Note that pybind11 comes with an example of building using scikit-build.
# (see also https://www.benjack.io/hybrid-python/c-packages-revisited/)
def customize_compiler_for_nvcc(compiler):
    compiler.src_extensions.append(".cu")

    if compiler.compiler_type == "unix":
        default_compile = compiler._compile
        default_compiler_so = compiler.compiler_so

        def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
            if os.path.splitext(src)[1] == ".cu":
                compiler.set_executable("compiler_so", "nvcc")
                postargs = extra_postargs["cuda"]
                # Display number of GPU registered used (in case of 701=cudaErrorLaunchOutOfResources):
                # postargs += ["-Xptxas", "-v"]
            elif "vendored" in src:
                postargs = extra_postargs["vendored-c++"]
            else:
                postargs = extra_postargs["c++"]

            default_compile(obj, src, ext, cc_args, postargs, pp_opts)
            compiler.compiler_so = default_compiler_so

        compiler._compile = _compile
    elif compiler.compiler_type == "msvc":
        compiler._cpp_extensions += [".cu"]
        compiler.initialize()

        nvcc = os.path.join(windows_cuda_path, "bin", "nvcc")

        default_compile = compiler.compile
        default_spawn = compiler.spawn
        default_cc = compiler.cc

        def compile(
            sources,
            output_dir=None,
            macros=None,
            include_dirs=None,
            debug=0,
            extra_preargs=None,
            extra_postargs=None,
            depends=None,
        ):
            cpp_sources = []
            vendored_cpp_sources = []
            cuda_sources = []
            for src in sources:
                if os.path.splitext(src)[1] == ".cu":
                    cuda_sources.append(src)
                elif "vendored" in src:
                    vendored_cpp_sources.append(src)
                else:
                    cpp_sources.append(src)

            cpp_objects = default_compile(
                cpp_sources,
                output_dir=output_dir,
                macros=macros,
                include_dirs=include_dirs,
                debug=debug,
                extra_preargs=extra_preargs,
                extra_postargs=extra_postargs["c++"],
                depends=depends,
            )

            vendored_cpp_objects = default_compile(
                vendored_cpp_sources,
                output_dir=output_dir,
                macros=macros,
                include_dirs=include_dirs,
                debug=debug,
                extra_preargs=extra_preargs,
                extra_postargs=extra_postargs["vendored-c++"],
                depends=depends,
            )

            cuda_objects = []
            if cuda_sources:
                compiler.cc = nvcc
                cuda_objects = default_compile(
                    cuda_sources,
                    output_dir=output_dir,
                    macros=macros,
                    include_dirs=include_dirs,
                    debug=debug,
                    extra_preargs=extra_preargs,
                    extra_postargs=extra_postargs["cuda"],
                    depends=depends,
                )
                compiler.cc = default_cc

            return cpp_objects + vendored_cpp_objects + cuda_objects

        def spawn(command):
            if command[0] == nvcc:
                patched_command = [nvcc]
                xcompiler = False
                for part in command[1:]:
                    if part == "-Xcompiler":
                        patched_command.append(part)
                        xcompiler = True
                    elif xcompiler:
                        patched_command.append(part)
                        xcompiler = False
                    elif part.startswith("-"):
                        patched_command.append(part)
                    elif part == "/c":
                        patched_command.append("-c")
                    elif part.startswith("/O"):
                        patched_command.append("-O" + part[2:])
                    elif part.startswith("/Tp"):
                        patched_command.append(part[3:])
                    elif part.startswith("/Fo"):
                        patched_command.append("-o")
                        patched_command.append(part[3:])
                    else:
                        print("WARNING: option passed by distutil to cl ignored when calling nvcc: " + part, file=sys.stderr)
                default_spawn(patched_command)
            else:
                default_spawn(command)

        compiler.compile = compile
        compiler.spawn = spawn
    else:
        assert False, f"Unsupported compiler type: {compiler.compiler_type}"


class custom_build_ext(setuptools.command.build_ext.build_ext):
    def build_extensions(self):
        customize_compiler_for_nvcc(self.compiler)
        setuptools.command.build_ext.build_ext.build_extensions(self)


def make_liblincs_extension():
    define_macros = [("NDEBUG", None), ("DOCTEST_CONFIG_DISABLE", None)]

    sources = []
    sources += glob.glob(os.path.join("lincs", "liblincs", "**", "*.c"), recursive=True)
    sources += glob.glob(os.path.join("lincs", "liblincs", "**", "*.cc"), recursive=True)
    sources += glob.glob(os.path.join("lincs", "liblincs", "**", "*.cpp"), recursive=True)

    # Non-standard: the dict is accessed in `customize_compiler_for_nvcc`
    # to get the standard form for `extra_compile_args`
    extra_compile_args = {}

    libraries = []
    include_dirs = []
    library_dirs = []
    extra_link_args=[]

    if os.environ.get("LINCS_DEV_FORBID_NVCC", "false") != "true" and shutil.which("nvcc") is not None:
        sources += glob.glob(os.path.join("lincs", "liblincs", "**", "*.cu"), recursive=True)
        libraries += ["cudart"]
        define_macros += [("LINCS_HAS_NVCC", None)]
        # @todo(Project management, later) Support several versions of CUDA?
        if sys.platform == "linux":
            include_dirs += ["/usr/local/cuda-12.1/targets/x86_64-linux/include"]
            library_dirs += ["/usr/local/cuda-12.1/targets/x86_64-linux/lib"]
            extra_compile_args["cuda"] = ["-std=c++17", "-Xcompiler", "-fopenmp,-fPIC,-Werror=switch"]
        elif sys.platform == "win32":
            include_dirs += [os.path.join(windows_cuda_path, "include")]
            library_dirs += [os.path.join(windows_cuda_path, r"lib\x64")]
            extra_compile_args["cuda"] = ["-std=c++17", "-Xcompiler", "/openmp,/MD"]
        else:
            assert False, f"Unsupported platform for CUDA: {sys.platform}"
    else:
        if os.environ.get("LINCS_DEV_FORCE_NVCC", "false") == "true":
            raise Exception("nvcc is not available but LINCS_DEV_FORCE_NVCC is true")
        else:
            print("WARNING: 'nvcc' was not found, lincs will be compiled without CUDA support", file=sys.stderr)

    try:
        chrones_dir = subprocess.run(
            ["chrones", "instrument", "c++", "header-location"], capture_output=True, universal_newlines=True, check=True
        ).stdout.strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        chrones_dir = None

    if os.environ.get("LINCS_DEV_FORBID_CHRONES", "false") != "true" and chrones_dir is not None:
        include_dirs += [chrones_dir]
        define_macros += [("LINCS_HAS_CHRONES", None)]
    else:
        if os.environ.get("LINCS_DEV_FORCE_CHRONES", "false") == "true":
            raise Exception("Chrones is not available but LINCS_DEV_FORCE_CHRONES is true")
        else:
            print("WARNING: 'chrones' was not found, lincs will be compiled without Chrones", file=sys.stderr)

    if sys.platform == "linux":
        extra_compile_args["c++"] = ["-std=c++17", "-Werror=switch", "-fopenmp"]
        extra_compile_args["vendored-c++"] = ["-std=c++17", "-Werror=switch", "-w", "-DQUIET", "-DNBUILD", "-DNCONTRACTS"]
        extra_link_args += ["-fopenmp"]
        libraries += [
            f"boost_python{sys.version_info.major}{sys.version_info.minor}",
            "ortools",
            # Weirdly required because of BoostPython:
            f"python{sys.version_info.major}.{sys.version_info.minor}{'m' if sys.hexversion < 0x03080000 else ''}",
        ]
        if os.environ.get("LINCS_DEV_COVERAGE", "false") == "true":
            extra_compile_args["c++"] += ["--coverage", "-O0"]
            extra_link_args += ["--coverage"]
    elif sys.platform == "win32":
        define_macros += [
            ("__WIN32", None),  # For Cadical inside EvalMaxSat
            ("_SILENCE_ALL_CXX17_DEPRECATION_WARNINGS", None),  # Silence a few warnings for OR-Tools
        ]
        extra_compile_args["c++"] = ["/std:c++17", "/openmp"]
        extra_compile_args["vendored-c++"] = ["/std:c++17", "-w", "-DQUIET", "-DNBUILD", "-DNCONTRACTS"]
        lincs_dependencies = os.environ.get("LINCS_DEV_DEPENDENCIES", os.path.join("c:", "lincs-deps"))
        include_dirs += [os.path.join(lincs_dependencies, "include")]
        library_dirs += [os.path.join(lincs_dependencies, "lib")]
        vc_version = os.environ.get("LINCS_DEV_VC_VERSION", "143")
        libraries += [
            f"boost_python{sys.version_info.major}{sys.version_info.minor}-vc{vc_version}-mt-x64-1_82",
            "ortools",
            f"python{sys.version_info.major}{sys.version_info.minor}",
        ]
    elif sys.platform == "darwin":
        extra_compile_args["c++"] = ["-std=c++17", "-Werror=switch", "-Xclang", "-fopenmp"]
        extra_compile_args["vendored-c++"] = ["-std=c++17", "-Werror=switch", "-w", "-DQUIET", "-DNBUILD", "-DNCONTRACTS"]
        extra_link_args += ["-lomp"]
        libraries += [
            f"boost_python{sys.version_info.major}{sys.version_info.minor}",
            "ortools",
        ]
    else:
        assert False, f"Unsupported platform: {sys.platform}"

    return setuptools.Extension(
        name="liblincs",
        sources=sources,
        libraries=libraries,
        define_macros=define_macros,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )

setuptools.setup(
    name="lincs",
    version=version,
    description="Learn and Infer Non Compensatory Sortings",
    license="LGPLv3",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/MICS-Lab/lincs",
    author="Vincent Jacques",
    author_email="vincent@vincent-jacques.net",
    install_requires=install_requires,
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "lincs = lincs.command_line_interface:main",
        ],
    },
    ext_modules=[make_liblincs_extension()],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    cmdclass={
        "build_ext": custom_build_ext,
    },
)



# Copyright 2023 Vincent Jacques

__version__ = "1.0.0"

# I/O
from liblincs import DataValidationException

from liblincs import Criterion, Category, Problem

from liblincs import SufficientCoalitions, Model
from liblincs import Alternative, Alternatives

# Generation (incl. misclassification)
from liblincs import BalancedAlternativesGenerationException
from liblincs import generate_classification_problem, generate_mrsort_classification_model, generate_classified_alternatives, misclassify_alternatives

# Classification
from liblincs import classify_alternatives

# Learning
from liblincs import LearningFailureException

# Learning - weights-profiles-breed
from liblincs import LearnMrsortByWeightsProfilesBreed
from liblincs import InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion
from liblincs import OptimizeWeightsUsingGlop, OptimizeWeightsUsingAlglib
from liblincs import ImproveProfilesWithAccuracyHeuristicOnCpu
try:
    from liblincs import ImproveProfilesWithAccuracyHeuristicOnGpu
    has_gpu = True
except ImportError:
    has_gpu = False
from liblincs import ReinitializeLeastAccurate
from liblincs import TerminateAtAccuracy
from liblincs import TerminateAfterSeconds, TerminateAfterSecondsWithoutProgress
from liblincs import TerminateAfterIterations, TerminateAfterIterationsWithoutProgress
from liblincs import TerminateWhenAny

# Learning - SAT by coalitions
from liblincs import LearnUcncsBySatByCoalitionsUsingMinisat

# Learning - SAT by separation
from liblincs import LearnUcncsBySatBySeparationUsingMinisat

# Learning - max-SAT by coalitions
from liblincs import LearnUcncsByMaxSatByCoalitionsUsingEvalmaxsat

# Learning - max-SAT by separation
from liblincs import LearnUcncsByMaxSatBySeparationUsingEvalmaxsat

# @todo(Feature, later) Accept learning and training set as Pandas DataFrame?
